/**
 * Copyright 2008, Daniel Molina Cabrera <danimolina@gmail.com>
 * 
 * This file is part of software Realea
 * 
 * Realea is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Realea is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "lscross.h"
#include <realea/common/distance.h>
#include <realea/common/random.h>
#include <realea/common/populationreal.h>
#include <cmath>

using namespace realea;

class LSCrossParams: public ILSParameters {
    public:
	tChromosomeReal *parents;
	tFitness *fitness;
	bool init;

	LSCrossParams(int num, int dim) {
	    parents = new tChromosomeReal[num];

	    for (int i = 0; i < num; i++) {
		parents[i].reserve(dim);
	    }
	    fitness = new tFitness [num];
	    init = false;
	}

	~LSCrossParams(void) {
	    delete[] parents;
	    delete[] fitness;
	}
};

LSCross::LSCross(double alpha) {
    m_alpha = alpha;
}

ILSParameters *LSCross::getInitOptions(tChromosomeReal &sol) {
    int size = m_pop->size();
    LSCrossParams *params=new LSCrossParams(3, sol.size());
    tIndividualReal *ind;
    int sample[size], pos;

    initSample(sample, size);

    // Create two parent random
    pos = m_random->getSample(sample, &size);
    ind = m_pop->getInd(pos);
    params->parents[1] = ind->sol();
    params->fitness[1] = ind->perf();
    pos = m_random->getSample(sample, &size);
    ind = m_pop->getInd(pos);
    params->parents[2] = ind->sol();
    params->fitness[2] = ind->perf();

    return params;
}

void crossoverSPX(Random *random, tChromosomeReal *parents,unsigned np, unsigned dim, tChromosomeReal &off) {
    tChromosomeReal o(dim);
    tChromosomeReal r(np);
    tChromosomeReal y(np),C(np);
    double u;
    unsigned d, i, E=1;

    for (d = 0; d < dim; d++) {
       o[d] = 0;

       for (i = 0; i < np; ++i) {
	    o[d] += parents[i][d];
	}
       o[d] /= np;
    }

    for (i = 0; i < np; i++) {
        u = random->rand();
	r[i] = pow(u, 1.0/(i+1));
    }

    for (d = 0; d < np; d++) {
	y[0] = parents[0][d];
	C[0] = 0;
	
	for (i = 1; i < np; i++) {
	    y[i] = o[d]+E*(parents[i][d]-o[d]);
	    C[i] = r[i-1]*(y[i-1]-y[i]+C[i-1]);
	}

	off[d] = y[np-1]+C[np-1];
    }
    
}

unsigned LSCross::apply(ILSParameters *opt, tChromosomeReal &sol, tFitness &fitness, unsigned maxeval) {
   LSCrossParams *p =  (LSCrossParams *) opt;
   unsigned numEval;
   int ndim = sol.size();

   if (!p->init) {
      p->parents[0] = sol;
      p->fitness[0] = fitness;
      p->init = true;
   }

   for (numEval = 0; numEval < maxeval && !m_running->isFinish(); ) {
        tChromosomeReal offspring(ndim);
	tFitness off_fitness;

	crossoverSPX(m_random, p->parents, 3, ndim, offspring);
	off_fitness = m_eval->eval(offspring);
	numEval++;

	if (m_problem->isBetter(off_fitness, p->fitness[0])) {
	    p->fitness[0] = off_fitness;
	    p->parents[0] = offspring;
	}
   }

   m_running->reset();
   sol = p->parents[0];
   fitness = p->fitness[0];
   return numEval;
}

unsigned LSCross::applyNeigh(tChromosomeReal &sol, PopulationReal *pop, tFitness &fitness, unsigned maxeval) {
   int ndim = sol.size();
   LSCrossParams *params =  new LSCrossParams(3, ndim);
   unsigned pos,numEval;
   double dist;
   tIndividualReal *ind;

   params->parents[0] = sol;
   params->fitness[0] = fitness;
   pos = getNeigh(sol, pop, 0);
   ind = m_pop->getInd(pos);
   params->parents[1] = ind->sol();
   params->fitness[1] = ind->perf();
   dist = distreal(sol, ind->sol());
   pos = getNeigh(sol, pop, dist);
   ind = m_pop->getInd(pos);
   params->parents[2] = ind->sol();
   params->fitness[2] = ind->perf();

   for (numEval = 0; numEval < maxeval && !m_running->isFinish(); ) {
        tChromosomeReal offspring(ndim);
	tFitness off_fitness;

	crossoverSPX(m_random, params->parents, 3, ndim, offspring);
	off_fitness = m_eval->eval(offspring);
	numEval++;

	if (m_problem->isBetter(off_fitness, params->fitness[0])) {
	    params->fitness[0] = off_fitness;
	    params->parents[0] = offspring;
	}
   }

   m_running->reset();
   sol = params->parents[0];
   fitness = params->fitness[0];
   return numEval;
}

