/**
 * Copyright 2008, Daniel Molina Cabrera <danimolina@gmail.com>
 * 
 * This file is part of software Realea
 * 
 * Realea is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Realea is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "malschains.h"
#include "selectlsimp.h"
#include <cassert>
#include <cstdio>
#include <ctime>
#include <cmath>

using namespace realea;
using namespace realea::internal;

void MALSChains::setEffortRatio(double ratio) {
	if (ratio == 1)
	   throw new string("MALSChains::effortRatio is not valide");

	m_effort = ratio;
}

MALSChains::MALSChains(IEAlgorithm *alg, ILocalSearch *ls) : Hybrid(alg,ls), m_memory(NULL) {
   m_effort = -1;
   m_nevalalg = m_nevalls = 0;
   m_restart = NULL;
   m_debug = false;
   m_select_improvement = NULL;
   m_disruption_size = 0;
}

void MALSChains::setDisruptionSize(double size) { 
   assert(size >= 0 && size < 1);
   m_disruption_size = size;
}


void MALSChains::setDebug(void) {
    m_debug = true;
}

MALSChains::~MALSChains(void) {
    if (m_memory) {
	delete m_memory;
    }

    if (m_restart) {
	delete m_restart;
    }

    if (m_select_improvement) {
	delete m_select_improvement;
    }

    if (m_initeval) {
	delete m_initeval;
    }
}

/**
 * Calculate the new frec to apply to obtain the indicated ratio.
 *
 * It is required because there are AEs that could not follow the frec indicated in the previous step
 *
 * @param nevalalg total number evaluation during the EA
 * @param nevalls total number evaluation during the LS
 * @param intensity LS intensity
 *
 * @return frec to use the EA to maintain the ratio
 */
unsigned calculateFrec(unsigned nevalalg, unsigned nevalls, unsigned intensity, double ratio) {

    assert(ratio > 0);

    double coc = ratio*(nevalalg + nevalls + intensity) - nevalalg;
    double div = (1-ratio);

    if (div == 0) {
	throw string("MALSChains::Ratio is too low");
    }

    return ((unsigned) floor(coc/div));
}

unsigned MALSChains::init(void) {
    initLs();

    m_initMaxEval = m_running->maxEval();
    unsigned neval = m_alg->init();

    if (m_select_improvement == NULL) {
	m_select_improvement = new SelectBestToImprove();
    }

    if (m_memory == NULL) {
	m_memory = new LSParametersMemory(m_alg->getPop()->size());
	m_alg->getPop()->setObserver(m_memory);
    }

    m_nevalalg = neval;
    m_nevalls = 0;
    return neval;
}

void MALSChains::setMaxEval(unsigned int maxeval) {
    assert(maxeval >= m_intensity);
    unsigned frec = calculateFrec(m_nevalalg, m_nevalls, m_intensity, m_effort);
    m_alg->setMaxEval(frec);
}

void MALSChains::setRestart(RestartStrategy *restart, int maxfailed) {

    if (m_restart) {
	delete m_restart;

    }

    m_restart = restart;
    m_maxfailed = maxfailed;
}

RunningPtr MALSChains::getRunning(void) {
    return m_alg->getRunning();
}

void MALSChains::setRunning(Running *running) {
    ProxyEA::setRunning(running);
    unsigned frec = calculateFrec(m_nevalalg, m_nevalls, m_intensity, m_effort);
    m_alg->setRunning(m_running->getSubRunning(frec));
}



void MALSChains::recoverIndividual(unsigned oldind, tGen *aind, unsigned size, tGen *aoptional, unsigned size_optional) {
    m_alg->recoverIndividual(oldind, aind, size-1, aoptional, size_optional);

    if (aind[size]) {
        m_alg->getPop()->getInd(oldind)->incremCount("non_improved");
    }

    if (aoptional != NULL) {
        IParallelLocalSearch *ls = (IParallelLocalSearch *) m_ls;
	ILSParameters *params = ls->recoverOptions(aoptional, size_optional);
	assert(m_memory);
	m_memory->store(oldind, params);
    }

}

void MALSChains::storeIndividual(tIndividualRealPtr ind, tGen **paind, unsigned *pmax, tGen **paoptional, unsigned *psize_optional) {
    tGen *asol, *aoptional_sol;
    tGen *asol_ma;
    unsigned size_sol, size_optional;
    tGen *aparams;
    unsigned size_param;
    unsigned size;

    m_alg->storeIndividual(ind, &asol, &size_sol, &aoptional_sol, &size_optional);
    assert(size_optional == 0 && aoptional_sol == NULL);

    asol_ma = new tGen[size_sol+1];
    copy(asol, asol+size_sol, asol_ma);
    delete[] asol;
    asol_ma[size_sol] = (ind->getCount("non_improved") > 0) ? 1 : 0;

    *paind = asol_ma;
    *pmax = size_sol+1;

    size_param = 0;
    
    if (m_memory) {
        unsigned posind = ind->getId();
        IParallelLocalSearch *ls = (IParallelLocalSearch *) m_ls;
	ILSParameters *params = m_memory->recover(posind);
	ls->storeOptions(params, &aparams, &size_param);
    }

    size = size_optional+size_param;
    assert(size > 0);

    *psize_optional = size_param;
    *paoptional = NULL;

    if (aoptional_sol != NULL || aparams != NULL) {
       *paoptional = new tGen[size];

       if (aoptional_sol != NULL) {
	  copy(aoptional_sol, aoptional_sol+size_optional, *paoptional);
	  delete[] aoptional_sol;
       }

       if (aparams != NULL) {
	  copy(aparams, aparams+size_param, *paoptional+size_optional);
	  delete[] aparams;
       }

    }
}

void MALSChains::setDif(bool debug, string ident, unsigned id, tFitness oldfit, tFitness newfit) {
   if (debug) {

      if (oldfit!= newfit) {
	 printf("%s[%2d]:\t%Le-> %Le\t\t%Le\n", ident.c_str(), id, oldfit, newfit, fabs(newfit-oldfit));
      }
//      else {
//	 printf("%s[%2d]:\t%Le\n", ident.c_str(), id, newfit);
//      }
   }

}

bool MALSChains::hasDiversity(PopulationReal *pop) {
    return true;
    double percen[5];
    pop->getPercentils(percen, 4);


	    printf("EA::Improvement: %e\t%e\t%e\t%e\t%e\n", percen[0], percen[1], 
			percen[2], percen[3], percen[4]);


    if (percen[2] == percen[4]) {
	return false;
    }
    else if (percen[1] == percen[3]) {
	return false;
    }
    else if (fabs((percen[0] - percen[2])/percen[2]) < 1e-3) {
	return false;
    }
    else {
	return true;    
    }
}

void MALSChains::setInitEval(IEval*eval) {
   Hybrid::setInitEval(eval);
   m_initeval = (IEvalInd *) new EvalRunning(eval, m_running);
}

void MALSChains::disturb(tChromosomeReal &sol) {
    DomainRealPtr domain = m_problem->getDomain();
    unsigned dim = domain->getDimension();
    double min, max;
 
    for (unsigned i = 0; i < dim; i++) {
        if (domain->canBeChanged(i)) {
	    domain->getValues(i, &min, &max);
	    sol[i] += m_disruption_size*m_random->randreal(-1,1)*(max-min);
	}
    }

    domain->clip(sol);
}

unsigned MALSChains::realApply(tChromosomeReal &bestsol, tFitness &bestfitness) {
    tIndividualReal *ind, *best;
    unsigned posind;
    tFitness oldfitness, fitness, fitness_alg;
    unsigned ndim = bestsol.size();
    tChromosomeReal sol(ndim), sol_alg(ndim);
    unsigned alg_failed;
    int lastpos_better = -1;
    tFitness lastpos_fitness;
    clock_t m_time_ls, m_time_alg, m_time_ma;
    clock_t clock_begin, m_time_ma_begin;
    
    PopulationReal *pop_alg = m_alg->getPop();
    deque<tIndividualReal*> ind_to_improve;
    tFitness improvement_alg, improvement_ls;

    m_time_ls = m_time_alg = m_time_ma = 0;

    sol=bestsol;
    fitness = bestfitness;
    improvement_alg = improvement_ls = 0;

    unsigned initMax = m_running->numEval();
    fitness_alg = pop_alg->getInd(pop_alg->getBest())->perf();

    alg_failed = 0;
    m_time_ma_begin = clock();

//    while ( (m_nevalalg+m_nevalls) < m_initMaxEval && !m_running->hasFoundOptime()) {
    while (!m_running->isFinish()) {
        tFitness old_fitness = fitness_alg;
	clock_begin = clock();
	m_nevalalg += m_alg->realApply(sol_alg, fitness_alg);
	m_time_alg += clock()-clock_begin;
	improvement_alg += fabs(fitness_alg-old_fitness);
	assert(fitness_alg <= old_fitness);

	if (fitness_alg == old_fitness) {
	    alg_failed++;
	}
	else {
	    alg_failed = 0;
	}

	if (m_debug) {
	    setDif(m_debug, "Alg", pop_alg->getBest(), old_fitness, fitness_alg);
	}

	// Check the optime
	if (m_running->isOptime(fitness_alg)) {
	    continue;
	}

	// Select the new individual to increm
	m_select_improvement->getIndsToImprove(pop_alg, ind_to_improve);

	if (ind_to_improve.size()!=0) {
	    // Select the individual to improve
	    posind = m_select_improvement->selectIndToImprove(ind_to_improve);
	}
	else {
	    // Choose the new one randomly
	    posind = m_random->randint(0, pop_alg->size()-1);
	}

	ind = pop_alg->getInd(posind);
	sol = ind->sol();

	ILSParameters *params = m_memory->recover(posind);
	bool recover = (params != NULL);
	
	if (params == NULL) {	
	    // Apply the LS to the best one with the rest of intensity
	    params = m_ls->getInitOptions(sol);
	}
	
	fitness = ind->perf();
	oldfitness = fitness;

	clock_begin= clock();
        m_nevalls += 
	    m_ls->apply(params, sol, fitness, m_intensity);
	m_time_ls += clock()-clock_begin;

	ind->incremCount("ls"); 
	improvement_ls += fabs(fitness-oldfitness);

	setDif(m_debug, "LS ", ind->getId(), oldfitness, fitness);

	if (m_problem->isBetter(fitness, fitness_alg)) {
	    fitness_alg = fitness;
	}

	if (!m_problem->isBetter(fitness, oldfitness)) {
	   ind->incremCount("non_improved");

	   if (recover) {
	      m_memory->remove(posind);
	   }
	   else {
	      delete params;
	   }

	   if (m_disruption_size > 0) {
	      unsigned pos_best = pop_alg->getBest();
	      tIndividualRealPtr ind_dis=NULL;

	      if (posind != pos_best) {
		 ind_dis = ind;
	      }
	      else {
		 if (lastpos_better >= 0) {
		 ind_dis = pop_alg->getInd(lastpos_better);

		 if (ind_dis->perf() != lastpos_fitness || pos_best == lastpos_better) {
		    ind_dis = NULL;
		 }
		}

		 lastpos_better = pos_best;
		 lastpos_fitness = pop_alg->getInd(lastpos_better)->perf();
	      }

	      if (ind_dis != NULL) {
		 tChromosomeReal sol = ind_dis->sol();
		 disturb(sol);
		 fitness = m_eval->eval(sol);
		 pop_alg->change(ind_dis->getId(), sol, fitness);
	      }

	   }

	   
	   if (m_disruption_size == 0 && ind_to_improve.empty() && m_restart != NULL && !m_running->isFinish()) {
		m_restart->apply(pop_alg, m_problem, m_initeval);
		
		if (m_debug) {
		    printf("Restart_NotImprovement\t: %Le\n", fitness);
		    printf("RunningRatio: %.0f\n", 100*m_running->ratio());
		}
	   }
	}
	else {
	    pop_alg->change(posind, sol, fitness);
	    m_memory->store(posind, params);
	}

	if (m_restart != NULL && m_maxfailed > 0) {
	     if (alg_failed > m_maxfailed && !hasDiversity(pop_alg) && !m_running->isFinish()) {
		m_restart->apply(pop_alg, m_problem, m_initeval);
		
		if (m_debug) {
		    printf("Restart_AlgFailed\t: %Le\n", fitness_alg);
		}
		alg_failed = 0;
	     }
	}

	fflush(stdout);
    }

    m_time_ma = clock()-m_time_ma_begin;

    if (m_debug) {
        double ratio_effort = ( (double) m_nevalalg)/(m_nevalalg+m_nevalls);
	printf("RatioEffort Alg/LS: [%.0f/%.0f]\n", 100*ratio_effort, 100*(1-ratio_effort));

        double ratio_alg = improvement_alg/(improvement_alg+improvement_ls);
	printf("RatioImprovement Alg/LS: [%.0f/%.0f]\n", 100*ratio_alg, 100*(1-ratio_alg));
    }
    
    double time_ms_alg, time_ms_ls, time_ms_ma;

    time_ms_alg = (m_time_alg*1000.0)/CLOCKS_PER_SEC;
    time_ms_ls = (m_time_ls*1000.0)/CLOCKS_PER_SEC;
    time_ms_ma = (m_time_ma*1000.0)/CLOCKS_PER_SEC;

    printf("Time[ALG]: %.2f\n", time_ms_alg);
    printf("Time[LS]: %.2f\n", time_ms_ls);
    printf("Time[MA]: %.2f\n", time_ms_ma);
    printf("Ratio[ALG/MA]: %.2f\n", 100*time_ms_alg/time_ms_ma);
    printf("Ratio[LS/MA]: %.2f\n", 100*time_ms_ls/time_ms_ma);
    printf("Ratio[(ALG+LS)/MA]: %.2f\n", 100*(time_ms_alg+time_ms_ls)/time_ms_ma);


    
    unsigned neval = m_running->numEval()-initMax;
    best = pop_alg->getInd(pop_alg->getBest());
    bestsol = best->sol();
    bestfitness = best->perf();
    m_running->reset();
    return neval;
}
